import os
from glob import glob
from models import model_baseline_2019_sampling, define_inference_subnet, define_sampling_subnet
import numpy as np
import tensorflow as tf
from tensorflow.keras.utils import to_categorical
from tensorflow.keras.callbacks import EarlyStopping, ModelCheckpoint, ReduceLROnPlateau, CSVLogger
from sklearn.model_selection import train_test_split
from sklearn.metrics import f1_score, accuracy_score, classification_report
from pylab import rcParams
rcParams['figure.figsize'] = 16, 12

import warnings
warnings.filterwarnings("ignore")

PP_DATA = "" #data_dir_path
WINDOW_SIZE = 100
bs = 64

def load_data():
    fnames = sorted(glob(os.path.join(PP_DATA, "*.npz")))
    total_fs = [f for f in fnames if f.split("/")[-1][:5]]
    total_data = {k: np.load(k) for k in total_fs}

    samples = total_data[os.path.join(PP_DATA,"SC4001E0.npz")]
   
    X = np.zeros((0, 3000, 1))
    y = []
    for fn in total_fs:
        samples = np.load(fn)
        X_data = samples["x"]
        X = np.concatenate((X, X_data), axis=0)
        y.extend(samples["y"])
    y = np.array(y)

    X_train, X_test, y_train, y_test = train_test_split(X,y,test_size=0.1,random_state=42)
    X_train, X_val, y_train, y_val = train_test_split(X_train,y_train,test_size=0.1,random_state=42)

    y_train = to_categorical(y_train)
    y_val = to_categorical(y_val)
    y_test = to_categorical(y_test)

    return X_train, y_train, X_test, y_test, X_val, y_val

def main():
    X_train, y_train, X_test, y_test, X_val, y_val = load_data()
    model_2019 = model_baseline_2019_sampling()
    checkpoint = ModelCheckpoint("model_cps", monitor='val_loss', verbose=1, save_best_only=True, mode='min')
    early = EarlyStopping(monitor="val_loss", mode="min", patience=10, verbose=1)
    redonplat = ReduceLROnPlateau(monitor="val_loss", mode="min", patience=5, verbose=2)
    csv_logger = CSVLogger('log_training.csv', append=True, separator=',')
    callbacks_list = [
        checkpoint,
        early,
        redonplat,
        csv_logger,
    ]

    hist = model_2019.fit(X_train, y_train, batch_size=64, epochs=50, validation_data=(X_val, y_val), callbacks=callbacks_list)
    y_pred = model_2019.predict(X_test, batch_size=bs)
    y_pred = np.array([np.argmax(s) for s in y_pred])

    model_2019.save_weights("deep_sleep_net_sampling.h5")
    model_2019.save("deep_sleep_net_sampling")

    y_test = np.array([np.argmax(s) for s in y_test])
    f1 = f1_score(y_test, y_pred, average="macro")
    report = classification_report(y_test, y_pred)
    accuracy_score(y_test, y_pred)


def compute_pruned_count(percent_pruning, layer):
    total = layer.output_shape[2]  
    nb_to_be_pruned = int(np.floor(percent_pruning * total))
    return nb_to_be_pruned

def get_weights_l1(layer):
    weights = layer.get_weights()[0]
    norms = []
    for i in range(weights.shape[0]):
        l1_norm = np.sum(abs(weights[i]))
        norms.append(l1_norm)
    return np.array(norms)

def set_weights_to_zero(model, layer, layer_id, perc):  
    weights_norm = get_weights_l1(layer)
    weights = layer.get_weights()[0] 
    n_pruned = compute_pruned_count(perc, layer)
    low_norm_weights = weights_norm.argsort()[:n_pruned]
    for i in range(0, low_norm_weights.shape[0]):
        x = low_norm_weights[i]
        weights[x] = 0
    
    model.layers[layer_id].set_weights([weights])
    model.compile(tf.keras.optimizers.Adam(0.001), tf.keras.losses.sparse_categorical_crossentropy, metrics=['acc'])  
    return model

def merge_models():
    model = model_baseline_2019_sampling()
    model = tf.keras.models.load_model("deep_sleep_net_sampling")
    model.summary()
    model.compile(tf.keras.optimizers.Adam(0.001), tf.keras.losses.sparse_categorical_crossentropy, metrics=['acc'])

    inference_subnet = define_inference_subnet()
    path = os.path.abspath('retrained_inference_net_deep_sleep')
    inference_subnet.load_weights(path)
    for i in range(1, len(inference_subnet.layers)):
        model.layers[i].set_weights(inference_subnet.layers[i].get_weights())
    
    model.save("retrained_full_model_deep_sleep")
    
    X_train, y_train, X_test, y_test, X_val, y_val = load_data()
    y_pred = model.predict(X_test, batch_size=bs)
    y_pred = np.array([np.argmax(s) for s in y_pred])
    y_test = np.array([np.argmax(s) for s in y_test])
    f1 = f1_score(y_test, y_pred, average="macro")
    report = classification_report(y_test, y_pred)
    print(report)
    accuracy_score(y_test, y_pred)
    return model

def retrain():
    X_train, y_train, X_test, y_test, X_val, y_val = load_data()
    model_2019 = model_baseline_2019_sampling()
    model_2019 = tf.keras.models.load_model("deep_sleep_net_sampling")

    y_pred = model_2019.predict(X_test, batch_size=bs)
    y_pred = np.array([np.argmax(s) for s in y_pred])
    y_test = np.array([np.argmax(s) for s in y_test])
    f1 = f1_score(y_test, y_pred, average="macro")
    report = classification_report(y_test, y_pred)
    accuracy_score(y_test, y_pred)

    layer_id = 1    
    subsampling_model = define_sampling_subnet()
    set_weights_to_zero(subsampling_model, subsampling_model.layers[layer_id], layer_id, perc=0.3)

    train_size = int(X_train.shape[0]/2)

    preds = np.zeros((X_train.shape[0], 3000,1))
    labels = np.zeros((X_train.shape[0], 5))
    k = 0
    for x,y in zip(X_train[:train_size,:,:], y_train[:train_size,]):
        X = np.expand_dims(x, 0)
        preds[k,:,:]= subsampling_model.predict(X)
        labels[k,:] = y
        k = k+1

    subsampling_model = define_sampling_subnet()
    set_weights_to_zero(subsampling_model, subsampling_model.layers[layer_id], layer_id, perc=0.7)

    for x,y in zip(X_train[train_size:,:,:], y_train[train_size:,]):
        X = np.expand_dims(x, 0)
        preds[k,:,:]= subsampling_model.predict(X)
        labels[k,:] = y
        k = k+1

    inference_model = define_inference_subnet()

    checkpoint = ModelCheckpoint('chk.h5', monitor='val_acc', verbose=1, save_best_only=True, mode='max')
    early = EarlyStopping(monitor="val_acc", mode="max", patience=20, verbose=1)
    redonplat = ReduceLROnPlateau(monitor="val_acc", mode="max", patience=5, verbose=2)
    callbacks_list = [checkpoint, early, redonplat]  

    inputs = np.array(preds)
    labels = np.array(labels)
    history = inference_model.fit(x=inputs, y=labels, epochs=100, verbose=2, steps_per_epoch=1000, validation_steps=300, callbacks=callbacks_list)
    path = os.path.abspath('retrained_inference_net_deep_sleep')
    inference_model.save_weights(path)
    inference_model.save("retrained_inference_net_deep_sleep")
