import os
import matplotlib.pyplot as plt
import pandas as pd
import tensorflow as tf
from tensorflow.keras import layers
from keras.layers import Conv2D, Dense, Flatten, MaxPooling2D, Dropout,BatchNormalization
from keras.callbacks import ModelCheckpoint, ReduceLROnPlateau, EarlyStopping
from SamplingLayer import SamplingLayer as SamplingLayer 
from LogMelSpectrogram import LogMelSpectrogram as LogMelSpectrogram
from audiomentations import Compose, AddGaussianNoise, TimeStretch

sr = 44100

def read_ESC_data(esc50_csv, base_data_path):
    pd_data = pd.read_csv(esc50_csv)
    filtered_pd = pd_data.loc[pd_data.esc10]

    classes = filtered_pd.iloc[:, 3].unique()
 
    my_classes = classes
    i = 0
    map_class_to_id = {}
    for c in my_classes:
        map_class_to_id[c] = i
        i = i+1
 
    nb_classes = i

    class_id = filtered_pd['category'].apply(lambda name: map_class_to_id[name])
    filtered_pd = filtered_pd.assign(target=class_id)

    full_path = filtered_pd['filename'].apply(lambda row: os.path.join(base_data_path, row))
    filtered_pd = filtered_pd.assign(filename=full_path)
    
    #Load the audio files
    filenames = filtered_pd['filename']
    targets = filtered_pd['target']
    folds = filtered_pd['fold']
    targets = tf.one_hot(targets, nb_classes)

    main_ds = tf.data.Dataset.from_tensor_slices((filenames, targets, folds))
    main_ds.element_spec

    main_ds = main_ds.map(load_wav_for_map)
    main_ds.element_spec

    #split the data
    cached_ds = main_ds.cache()
    train_ds = cached_ds.filter(lambda embedding, label, fold: fold < 4)
    val_ds = cached_ds.filter(lambda embedding, label, fold: fold == 4)
    test_ds = cached_ds.filter(lambda embedding, label, fold: fold == 5)

    # remove the folds column now that it's not needed anymore
    remove_fold_column = lambda embedding, label, fold: (embedding, label)

    train_ds = train_ds.map(remove_fold_column)
    val_ds = val_ds.map(remove_fold_column)
    test_ds = test_ds.map(remove_fold_column)

    augumented_train_ds = augment_audio_dataset(train_ds)
    combined_dataset = train_ds.concatenate(augumented_train_ds)

    train_ds = combined_dataset
    train_ds = train_ds.cache().shuffle(1000).batch(64).prefetch(tf.data.AUTOTUNE)
    val_ds = val_ds.cache().batch(64).prefetch(tf.data.AUTOTUNE)
    test_ds = test_ds.cache().batch(64).prefetch(tf.data.AUTOTUNE)

    num_labels = len(my_classes)

    return train_ds, val_ds, test_ds, num_labels

@tf.function
def load_wav_mono(filename):
    file_contents = tf.io.read_file(filename)
    wav, sample_rate = tf.audio.decode_wav(
          file_contents,
          desired_channels=1)
    wav = tf.squeeze(wav, axis=-1)
    return wav

def load_wav_for_map(filename, label, fold):
  return load_wav_mono(filename), label, fold

def apply_pipeline(y, sr):
    augmentations_pipeline = Compose(
    [
        AddGaussianNoise(min_amplitude=0.01, max_amplitude=0.15, p=0),
        TimeStretch(min_rate=.75, max_rate=1.25,p=1),
    ]
    )
    shifted = augmentations_pipeline(y, sr)
    return shifted

@tf.function
def tf_apply_pipeline(audio, label):
    augmented_feature = tf.numpy_function(apply_pipeline, inp=[audio, sr], Tout=tf.float32, name="apply_pipeline")
    return augmented_feature, label

def augment_audio_dataset(dataset: tf.data.Dataset):
    dataset = dataset.map(tf_apply_pipeline)
    return dataset

def create_model(dim = 220500, input_shape=(220500,), num_labels=None):
    model = tf.keras.models.Sequential()

    model.add(layers.Input(shape=input_shape))

    model.add(SamplingLayer(input_dim=dim))
    model.add(LogMelSpectrogram(sample_rate=44100, fft_size=2048, hop_size=1014, n_mels=128, f_min=0.0, f_max=None))
    model.add(layers.Permute((2,1,3)))
    model.add(BatchNormalization(axis=1, input_shape=input_shape))

    model.add(Conv2D(64, kernel_size=3, activation="relu", padding='same'))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Conv2D(128, kernel_size=3, activation="relu", padding='same'))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Conv2D(256, kernel_size=3, activation="relu", padding='same'))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Flatten())
    model.add(Dense(256, activation = 'relu'))
    model.add(Dropout(0.5))
    model.add(Dense(num_labels, activation="softmax"))

    # compile the model
    model.compile(optimizer='adam',
                    loss='categorical_crossentropy',
                    metrics=['accuracy'])
    model.summary()
    return model

def train_model(model, train_ds, val_ds, test_ds):
    opt = tf.keras.optimizers.Adam(lr=0.001)
    model.compile(loss='categorical_crossentropy',
                optimizer=opt,
                metrics=['acc'])
    model_dir = "./models"
    if not os.path.exists(model_dir):
        os.mkdir(model_dir)

    es_cb = EarlyStopping(monitor='val_loss', patience=10, verbose=1, mode='auto')
    rlronp = ReduceLROnPlateau(monitor="val_acc",factor=0.5, patience=2, verbose=1)

    chkpt = "esc50-{epoch:02d}-{val_acc:.2f}.hdf5"

    cp_cb = ModelCheckpoint(filepath = chkpt, monitor='val_loss', verbose=1, save_best_only=True, mode='auto')

    callbacks_list=[es_cb, cp_cb, rlronp]
    history = model.fit(
        train_ds,
        validation_data=val_ds,
        epochs=400,
        callbacks=[callbacks_list],
        verbose=2
    )

    plt.close('all')
    metrics = history.history
    plt.plot(history.epoch, metrics['loss'], metrics['val_loss'])
    plt.legend(['loss', 'val_loss'])
    plt.show()
    plt.savefig("training_loss_esc50")


def main():
    train_ds, val_ds, test_ds, num_labels = read_ESC_data()
    m = create_model(dim = 220500, input_shape=(220500,), num_labels=num_labels)
    trained_model = train_model(m, train_ds, val_ds, test_ds)


