import os
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import tensorflow as tf
import tensorflow_io as tfio
from tensorflow.keras import layers
from keras.layers import Conv2D, Dense, Flatten, MaxPooling2D, Permute,Activation, Resizing, GlobalAveragePooling2D, Dropout,BatchNormalization, Add
from keras.callbacks import ModelCheckpoint, ReduceLROnPlateau, EarlyStopping
from kapre.time_frequency import Melspectrogram
from kapre.utils import Normalization2D
from SamplingLayer import SamplingLayer as SamplingLayer 
from LogMelSpectrogram import LogMelSpectrogram as LogMelSpectrogram
from audiomentations import Compose, AddGaussianNoise, TimeStretch

sr = 44100

def read_ESC_data():
    esc50_csv = './datasets/ESC-50-master/ESC-50-master/meta/esc50.csv'
    base_data_path = './datasets/ESC-50-master/ESC-50-master/audio/'

    pd_data = pd.read_csv(esc50_csv)
    filtered_pd = pd_data.loc[pd_data.esc10]
    print(filtered_pd.head())

    classes = filtered_pd.iloc[:, 3].unique()
    print(classes)

    my_classes = classes
    i = 0
    map_class_to_id = {}
    for c in my_classes:
        map_class_to_id[c] = i
        i = i+1
    print(map_class_to_id)

    nb_classes = i

    class_id = filtered_pd['category'].apply(lambda name: map_class_to_id[name])
    filtered_pd = filtered_pd.assign(target=class_id)

    full_path = filtered_pd['filename'].apply(lambda row: os.path.join(base_data_path, row))
    filtered_pd = filtered_pd.assign(filename=full_path)
    
    #Load the audio files
    filenames = filtered_pd['filename']
    targets = filtered_pd['target']
    folds = filtered_pd['fold']
    targets = tf.one_hot(targets, nb_classes)

    main_ds = tf.data.Dataset.from_tensor_slices((filenames, targets, folds))
    main_ds.element_spec

    main_ds = main_ds.map(load_wav_for_map)
    main_ds.element_spec

    #split the data
    cached_ds = main_ds.cache()
    train_ds = cached_ds.filter(lambda embedding, label, fold: fold < 4)
    val_ds = cached_ds.filter(lambda embedding, label, fold: fold == 4)
    test_ds = cached_ds.filter(lambda embedding, label, fold: fold == 5)

    # remove the folds column now that it's not needed anymore
    remove_fold_column = lambda embedding, label, fold: (embedding, label)

    train_ds = train_ds.map(remove_fold_column)
    val_ds = val_ds.map(remove_fold_column)
    test_ds = test_ds.map(remove_fold_column)

    augumented_train_ds = augment_audio_dataset(train_ds)
    combined_dataset = train_ds.concatenate(augumented_train_ds)

    train_ds = combined_dataset
    train_ds = train_ds.cache().shuffle(1000).batch(64).prefetch(tf.data.AUTOTUNE)
    val_ds = val_ds.cache().batch(64).prefetch(tf.data.AUTOTUNE)
    test_ds = test_ds.cache().batch(64).prefetch(tf.data.AUTOTUNE)

    num_labels = len(my_classes)
    print(num_labels)
    return train_ds, val_ds, test_ds, num_labels

@tf.function
def load_wav_mono(filename):
    file_contents = tf.io.read_file(filename)
    wav, sample_rate = tf.audio.decode_wav(
          file_contents,
          desired_channels=1)
    wav = tf.squeeze(wav, axis=-1)
    return wav

def load_wav_for_map(filename, label, fold):
  return load_wav_mono(filename), label, fold

def apply_pipeline(y, sr):
    augmentations_pipeline = Compose(
    [
        AddGaussianNoise(min_amplitude=0.01, max_amplitude=0.15, p=0),
        TimeStretch(min_rate=.75, max_rate=1.25,p=1),
    ]
    )
    shifted = augmentations_pipeline(y, sr)
    return shifted

@tf.function
def tf_apply_pipeline(audio, label):
    augmented_feature = tf.numpy_function(apply_pipeline, inp=[audio, sr], Tout=tf.float32, name="apply_pipeline")
    return augmented_feature, label

def augment_audio_dataset(dataset: tf.data.Dataset):
    dataset = dataset.map(tf_apply_pipeline)
    return dataset

def compute_pruned_count(percent_pruning, layer):
    total = layer.output_shape[1]
    nb_to_be_pruned = int(np.floor(percent_pruning * total))
    return nb_to_be_pruned

def get_weights_l1_NEW(layer):
    weights = layer.get_weights()[0]
    norms = abs(weights)
    return norms

def set_weights_to_zero2(model, layer, layer_id, perc):
    weights_norm = get_weights_l1_NEW(layer)
    weights = layer.get_weights()[0] 
    n_pruned = compute_pruned_count(perc, layer)
    print("n_pruned",n_pruned)
    weights_norm = np.reshape(weights_norm,(weights_norm.shape[0]*weights_norm.shape[1]))
    min_indices = np.argpartition(weights_norm, n_pruned-1)[:n_pruned]
    weights_norm[min_indices] = 0
    weights_norm = np.reshape(weights_norm,(weights.shape[0],weights.shape[1]))
    
    mask = np.ones((weights.shape[0],weights.shape[1])).astype(bool)
    mask = np.logical_and(mask, weights_norm == 0)
    weights[mask] = 0
    model.layers[layer_id].set_weights([weights])
    model.compile(tf.keras.optimizers.Adam, tf.keras.losses.categorical_crossentropy, metrics=['accuracy'])  
    return model


def set_weights_to_zero(model, layer, layer_id, perc):  
    weights = layer.get_weights()[0] 
    weights_norm = abs(weights)
    #weights_norm = get_weights_l1(layer)
    n_pruned = compute_pruned_count(perc, layer)
    print("percentage pruned ",perc)
    low_norm_weights = weights_norm.argsort()[:n_pruned]
    weights[low_norm_weights] = 0
    # for i in range(0, low_norm_weights.shape[0]):
    #     x = low_norm_weights[i]
    #     weights[x] = 0
    
    model.layers[layer_id].set_weights([weights])
    return model

def compute_weights_to_zero(model, layer, perc):
    weights_norm = get_weights_l1_NEW(layer)
    weights = layer.get_weights()[0] 
    n_pruned = compute_pruned_count(perc, layer)
    print("n_pruned",n_pruned)
    weights_norm = np.reshape(weights_norm,(weights_norm.shape[0]*weights_norm.shape[1]))
    min_indices = np.argpartition(weights_norm, n_pruned-1)[:n_pruned]
    weights_norm[min_indices] = 0
    weights_norm = np.reshape(weights_norm,(weights.shape[0],weights.shape[1]))
    
    mask = np.ones((weights.shape[0],weights.shape[1])).astype(bool)
    mask = np.logical_and(mask, weights_norm == 0)
    weights[mask] = 0
    np.savetxt("pruned_weights"+str(np.round(perc,2))+".csv", weights, delimiter=",")
    return [weights]


def define_sampling_subnet(dim = 220500, input_shape=(220500,),):
    model = tf.keras.models.Sequential()
    model.add(layers.Input(shape=input_shape))
    model.add(SamplingLayer(input_dim=dim))
    model.compile(optimizer='adam',
                    loss='categorical_crossentropy',
                    metrics=['accuracy'])
    model.summary()
    return model


def create_model(dim = 220500, input_shape=(220500,), num_labels=None):
    model = tf.keras.models.Sequential()

    model.add(layers.Input(shape=input_shape))

    model.add(SamplingLayer(input_dim=dim))
    model.add(LogMelSpectrogram(sample_rate=44100, fft_size=2048, hop_size=1014, n_mels=128, f_min=0.0, f_max=None))
    model.add(layers.Permute((2,1,3)))
    model.add(BatchNormalization(axis=1, input_shape=input_shape))

    model.add(Conv2D(64, kernel_size=3, activation="relu", padding='same'))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Conv2D(64, kernel_size=3, activation="relu", padding='same'))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Conv2D(128, kernel_size=3, activation="relu", padding='same'))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Conv2D(128, kernel_size=3, activation="relu", padding='same'))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Conv2D(256, kernel_size=3, activation="relu", padding='same'))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Conv2D(256, kernel_size=3, activation="relu", padding='same'))
    model.add(MaxPooling2D(pool_size=(2, 2)))
    model.add(Flatten())
    model.add(Dense(256, activation = 'relu'))
    model.add(Dropout(0.5))
    model.add(Dense(num_labels, activation="softmax"))

    # compile the model
    model.compile(optimizer='adam',
                    loss='categorical_crossentropy',
                    metrics=['accuracy'])
    model.summary()
    return model


def make_predictions(model, trainds):
    column_names = list(range(220500))
    for i in range(len(column_names)):
        column_names[i] = str(i+1)

    column_names2 = list(range(10))
    for i in range(len(column_names2)):
        column_names2[i] = str(i+1)
 
    labels = np.concatenate([np.expand_dims(y,axis=0) for x, y in trainds], axis=0)
    
    s = np.empty((0, 220500))
    for x, y in trainds:
        predictions = model(x, training=False)
        #predictions = np.concatenate([predictions], axis=1)
        p = np.vstack((s, np.squeeze(np.array(predictions))))
        s = p
    print(p.shape)

    dataframe1 = pd.DataFrame(labels, columns=column_names2)
    dataframe2 = pd.DataFrame(p, columns=column_names)

    dataframe1.to_csv('labels.csv', mode='a', header = False, index = None)
    dataframe2.to_csv('samples.csv', mode='a', header = False, index = None)



def main():
    train_ds, val_ds, test_ds, num_labels = read_ESC_data()
    saved_model =  tf.keras.models.load_model('mobilenet_model_sampling_esc') 
    saved_model.summary()
    weights = saved_model.layers[0].get_weights()[0]
    
    layer_id = 0
    subsampling_model = define_sampling_subnet()
    subsampling_model = set_weights_to_zero(subsampling_model, saved_model.layers[layer_id], layer_id, 0.5) 
  
    l = []
    l.append(weights)
    subsampling_model.layers[0].set_weights(l)    
    
    perc = [0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9]
    for x,y in train_ds:
        for percentage in perc:
            print(percentage)
            model_pruned = set_weights_to_zero(subsampling_model, saved_model.layers[layer_id], layer_id, percentage) 
            mini_train_dataset = tf.data.Dataset.from_tensor_slices((x, y))
            make_predictions(model_pruned, trainds=mini_train_dataset)

#main()


