import numpy as np
import tensorflow as tf
import os
from tensorflow.keras.layers import Dropout,Flatten, Dense, Conv2D, Reshape, BatchNormalization, MaxPooling2D, Activation
from tensorflow.keras.datasets import mnist
from keras.utils.np_utils import to_categorical  
from keras.models import Sequential
from tensorflow.keras import optimizers
from SamplingLayer import SamplingLayer as SamplingLayer 
from train_CSDL_MNIST import createModel_MNIST
from keras.models import load_model  
from tensorflow.keras.callbacks import ModelCheckpoint, EarlyStopping, ReduceLROnPlateau

seed = 42
tf.random.set_seed(seed)
np.random.seed(seed)

def compute_pruned_count(percent_pruning, layer):
    total = layer.output_shape[1]*layer.output_shape[2]  
    nb_to_be_pruned = int(np.floor(percent_pruning * total))
    return nb_to_be_pruned

def get_weights_l1(layer):
    weights = layer.get_weights()[0]
    norms = abs(weights)
    return norms

def compute_weights_to_zero(layer, perc):
    weights_norm = get_weights_l1(layer)
    weights = layer.get_weights()[0] 
    n_pruned = compute_pruned_count(perc, layer)
    weights_norm = np.reshape(weights_norm,(weights_norm.shape[0]*weights_norm.shape[1]))
    min_indices = np.argpartition(weights_norm, n_pruned-1)[:n_pruned]
    weights_norm[min_indices] = 0
    weights_norm = np.reshape(weights_norm,(weights.shape[0],weights.shape[1]))
    mask = np.ones((weights.shape[0],weights.shape[1])).astype(bool)
    mask = np.logical_and(mask, weights_norm == 0)
    weights[mask] = 0
    return [weights]

def define_sampling_subnet(in_shape=(28,28), learningrate=0.0002):
    model = Sequential()
    model.add(tf.keras.layers.Input(shape=in_shape))
    model.add(SamplingLayer(input_dim=(28,28)))
    model.summary()
    opt = optimizers.Adam(learning_rate=learningrate, beta_1=0.5)
    model.compile(loss='categorical_crossentropy', optimizer=opt, metrics=['accuracy'])
    return model

def define_inference_subnet(in_shape=(28,28), learningrate=0.0002):
    model = Sequential()
    model.add(tf.keras.layers.Input(shape=in_shape))
    model.add(Reshape(target_shape=(28,28,1)))
    model.add(Conv2D(32, (3, 3), input_shape=(28,28,1)))
    model.add(Activation('relu'))
    BatchNormalization(axis=-1)
    model.add(Conv2D(32, (3, 3)))
    model.add(Activation('relu'))
    model.add(MaxPooling2D(pool_size=(2,2)))
    BatchNormalization(axis=-1)
    model.add(Conv2D(64,(3, 3)))
    model.add(Activation('relu'))
    BatchNormalization(axis=-1)
    model.add(Conv2D(64, (3, 3)))
    model.add(Activation('relu'))
    model.add(MaxPooling2D(pool_size=(2,2)))
    model.add(Flatten())
    BatchNormalization()
    model.add(Dense(512))
    model.add(Activation('relu'))
    BatchNormalization()
    model.add(Dropout(0.2))
    model.add(Dense(10))
    model.add(Activation('softmax'))
    model.summary()
    opt = optimizers.Adam(learning_rate=learningrate, beta_1=0.5)
    model.compile(loss='categorical_crossentropy', optimizer=opt, metrics=['accuracy'])
    return model

def retrain_model(path_to_pretrained_model):
    (X_train, y_train), (X_test, y_test) = mnist.load_data()

    X_train = X_train.astype('float32')
    X_test = X_test.astype('float32')

    y_train = to_categorical(y_train, num_classes=10)
    y_test = to_categorical(y_test, num_classes=10)

    X_train = X_train.astype('float32')
    X_test = X_test.astype('float32')

    X_train /= 255
    X_test /= 255

    # load pretrained model
    model = createModel_MNIST(in_shape=(28,28), learningrate=0.00002) 
    model = load_model(path_to_pretrained_model)
    model.compile(loss='categorical_crossentropy', optimizer=tf.keras.optimizers.Adam(learning_rate=0.00002), metrics=['accuracy'])
    
    layer_id = 0
    weights = model.layers[layer_id].get_weights()[0]  
    subsampling_model = define_sampling_subnet()
    l = []
    l.append(weights)
    subsampling_model.layers[0].set_weights(l)  

    inference_model = define_inference_subnet()
    perc = [0.6, 0.7, 0.8, 0.91, 0.93, 0.95, 0.97, 0.98] #list of sampling rates

    nb = 36000
    X_val = X_train[-nb:]
    y_val = y_train[-nb:]
    predictions = np.zeros_like(X_val)
    labels = np.zeros_like(y_val)

    k = 0
    for image, label in zip(X_val, y_val):
        percentage = perc[np.random.randint(len(perc), size=1)[0]]
        weights = compute_weights_to_zero(model, model.layers[layer_id],  percentage)    
        subsampling_model.layers[0].set_weights(weights)  
        predictions[k,:,:] = subsampling_model.predict(image)
        labels[k,:] = label
        k = k+1

    filepath = os.path.join(os.path.dirname(__file__),'checkpoints')
    checkpoint = ModelCheckpoint(filepath, monitor='val_accuracy', verbose=1, save_best_only=True, mode='max')
    early = EarlyStopping(monitor="val_loss", patience=25, verbose=1)
    redonplat = ReduceLROnPlateau(monitor='val_loss', factor=0.5, patience=5, min_lr=1e-7, cooldown=20, verbose=1)
    callbacks_list = [checkpoint, early, redonplat]  
    
    history = inference_model.fit(predictions, labels, callbacks=callbacks_list, shuffle=True, verbose=2, epochs=100, validation_split=0.2)
    path = os.path.abspath('inference_subnet_multiple_rates')
    inference_model.save_weights(path)
    




