import os
import pathlib
import numpy as np
import tensorflow as tf
from speech_functions import load_data_batches, get_size_dataset
from models import define_model, define_inference_subnet, test_model

seed = 42
tf.random.set_seed(seed)
np.random.seed(seed)



def assemble_model(sampling_weights_path, num_labels, test_ds):
     
    model = define_model(num_labels=8, input_shape=(16000,1), dim=16000)
    model.compile(
        optimizer=tf.keras.optimizers.Adam(learning_rate=0.001),
        loss=tf.keras.losses.SparseCategoricalCrossentropy(from_logits=False),
        metrics=['sparse_categorical_accuracy'], 
    )   
    
    data = np.loadtxt(sampling_weights_path)
    w_shape = np.array(model.layers[1].get_weights()).shape
    data = np.reshape(data, w_shape)
    model.layers[1].set_weights(data)
    
    inference_subnet = define_inference_subnet(test_ds)
    inference_path = os.path.abspath('inference_subnet')
    inference_subnet.load_weights(inference_path)
    inference_subnet.compile(
        optimizer=tf.keras.optimizers.Adam(learning_rate=0.001),
        loss=tf.keras.losses.SparseCategoricalCrossentropy(from_logits=False),
        metrics=['sparse_categorical_accuracy'], 
    )  
       
    i = 2
    for j in range(1,len(inference_subnet.layers)):
        model.layers[i].set_weights(inference_subnet.layers[j].get_weights())
        i = i+1

    path = os.path.abspath('model_retrained_weights')
    model.save_weights(path)
    model.save('model_retrained')
    return model

def compute_pruned_count(percent_pruning, layer):
    total = layer.output_shape[1]
    nb_to_be_pruned = int(np.floor(percent_pruning * total))
    return nb_to_be_pruned


def prune_weights(layer, perc):  
    weights = layer.get_weights()[0] 
    weights_norm = abs(weights)
    n_pruned = compute_pruned_count(perc, layer)
    low_norm_weights = weights_norm.argsort()[:n_pruned]
    low_norm_weights = weights_norm.argsort()[:n_pruned]
    weights[low_norm_weights] = 0  
    return weights

def assemble_sparse_model(num_labels, test_ds, perc, filename, sampling_weights_path):
    model = define_model(num_labels=num_labels, input_shape=(16000,), dim=16000)
    model.compile(
        optimizer=tf.keras.optimizers.Adam(learning_rate=0.001),
        loss=tf.keras.losses.SparseCategoricalCrossentropy(from_logits=False),
        metrics=['sparse_categorical_accuracy'], 
    )   
    data = np.loadtxt(sampling_weights_path)
    w_shape = np.array(model.layers[1].get_weights()).shape
    data = np.reshape(data, w_shape)
    model.layers[1].set_weights(data)
    weights = prune_weights(model.layers[1], perc)
    model.layers[1].set_weights([weights])

    inference_subnet = define_inference_subnet(test_ds)
    inference_path = os.path.abspath('inference_subnet')
    inference_subnet.load_weights(inference_path)
    inference_subnet.compile(
        optimizer=tf.keras.optimizers.Adam(learning_rate=0.001),
        loss=tf.keras.losses.SparseCategoricalCrossentropy(from_logits=False),
        metrics=['sparse_categorical_accuracy'], 
    )  
       
    i = 2
    for j in range(1,len(inference_subnet.layers)):
        model.layers[i].set_weights(inference_subnet.layers[j].get_weights())
        i = i+1

    test_model(model, test_ds)
    model.save(filename)
    return model

def main():  
    sampling_weights_path =  os.path.abspath('weights.csv')

    # Load data
    train_ds, val_ds, test_ds, input_shape, wave_orig, commands = load_data_batches()
    train_data_size = get_size_dataset(train_ds)
    num_labels = len(commands)  

    for j in range(0, 10):
            perc = j/10
            filename ="model"+str(perc)
            model = assemble_sparse_model(num_labels, test_ds, perc, filename, sampling_weights_path)
            model = tf.keras.models.load_model(filename)
            test_model(model,test_ds)


