import pathlib
import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import layers
from speech_functions import make_spec_ds
from tensorflow.keras import backend as K
from kapre.time_frequency import Melspectrogram
from SamplingLayer import SamplingLayer


def define_model(train_ds, num_labels=8, input_shape=(16000,), dim=16000):
    norm_layer = layers.Normalization()
    train_spectrogram_ds = make_spec_ds(train_ds)
    norm_layer.adapt(data=train_spectrogram_ds.map(map_func=lambda spec, label: spec))

    inputs = layers.Input(shape=input_shape)
    sampling_layer = SamplingLayer(input_dim=dim)(inputs)
    x = layers.Reshape((1,-1))(sampling_layer)
    spectrogram = Melspectrogram(n_dft=1024, n_hop=128, input_shape=(1, dim),
                       padding='same', sr=dim, n_mels=80,
                       fmin=40.0, fmax=dim / 2, power_melgram=1.0,
                       return_decibel_melgram=True, trainable_fb=False,
                       trainable_kernel=False, trainable=False,
                       name='mel_stft')(x)
    spectrogram.trainable = False
    norm = norm_layer(spectrogram)

    x = layers.Permute((2, 1, 3))(norm)

    x = layers.Conv2D(10, (5, 1), activation='relu', padding='same')(x)
    x = layers.BatchNormalization()(x)
    x = layers.Conv2D(1, (5, 1), activation='relu', padding='same')(x)
    x = layers.BatchNormalization()(x)

    x = layers.Lambda(lambda q: K.squeeze(q, -1), name='squeeze_last_dim')(x)

    x = layers.Bidirectional(layers.LSTM(64, return_sequences=True))(x) 
    x = layers.Bidirectional(layers.LSTM(64, return_sequences=True))(x) 

    xFirst = layers.Lambda(lambda q: q[:, -1])(x)  # [b_s, vec_dim]
    query = layers.Dense(128)(xFirst)

    # dot product attention
    attScores = layers.Dot(axes=[1, 2])([query, x])
    attScores = layers.Softmax(name='attSoftmax')(attScores)  # [b_s, seq_len]

    # rescale sequence
    attVector = layers.Dot(axes=[1, 1])([attScores, x])  # [b_s, vec_dim]

    x = layers.Dense(64, activation='relu')(attVector)
    x = layers.Dense(32)(x)

    output = layers.Dense(num_labels, activation='softmax', name='output')(x)
    model =  tf.keras.Model(inputs=[inputs], outputs=[output])
    model.summary()    
    return model


def define_model_NO_SAMPLING(num_labels=8, input_shape=(16000,), dim=16000):
    norm_layer = layers.Normalization()
    train_spectrogram_ds = make_spec_ds(train_ds)
    norm_layer.adapt(data=train_spectrogram_ds.map(map_func=lambda spec, label: spec))

    inputs = layers.Input(shape=input_shape)

    x = layers.Reshape((1,-1))(inputs)
    spectrogram = Melspectrogram(n_dft=1024, n_hop=128, input_shape=(1, dim),
                       padding='same', sr=dim, n_mels=80,
                       fmin=40.0, fmax=dim / 2, power_melgram=1.0,
                       return_decibel_melgram=True, trainable_fb=False,
                       trainable_kernel=False, trainable=False,
                       name='mel_stft')(x)
    spectrogram.trainable = False
    norm = norm_layer(spectrogram)

    x = layers.Permute((2, 1, 3))(norm)

    x = layers.Conv2D(10, (5, 1), activation='relu', padding='same')(x)
    x = layers.BatchNormalization()(x)
    x = layers.Conv2D(1, (5, 1), activation='relu', padding='same')(x)
    x = layers.BatchNormalization()(x)

    x = layers.Lambda(lambda q: K.squeeze(q, -1), name='squeeze_last_dim')(x)

    x = layers.Bidirectional(layers.LSTM(64, return_sequences=True))(x) 
    x = layers.Bidirectional(layers.LSTM(64, return_sequences=True))(x) 

    xFirst = layers.Lambda(lambda q: q[:, -1])(x)  # [b_s, vec_dim]
    query = layers.Dense(128)(xFirst)

    # dot product attention
    attScores = layers.Dot(axes=[1, 2])([query, x])
    attScores = layers.Softmax(name='attSoftmax')(attScores)  # [b_s, seq_len]

    # rescale sequence
    attVector = layers.Dot(axes=[1, 1])([attScores, x])  # [b_s, vec_dim]

    x = layers.Dense(64, activation='relu')(attVector)
    x = layers.Dense(32)(x)

    output = layers.Dense(num_labels, activation='softmax', name='output')(x)
    model =  tf.keras.Model(inputs=[inputs], outputs=[output])
    model.summary()    
    return model

def define_sampling_subnet(input_shape=(16000,), dim=16000):
    inputs = layers.Input(shape=input_shape)
    sampling_layer = SamplingLayer(input_dim=dim)(inputs)
    model = tf.keras.Model(inputs=[inputs], outputs=[sampling_layer])
    model.summary()  
    return model

def define_inference_subnet(train_ds, num_labels=8, input_shape=(16000,), dim=16000):
    norm_layer = layers.Normalization()
    train_spectrogram_ds = make_spec_ds(train_ds)
    norm_layer.adapt(data=train_spectrogram_ds.map(map_func=lambda spec, label: spec))

    inputs = layers.Input(shape=input_shape)
    x = layers.Reshape((1,-1))(inputs)
    spectrogram = Melspectrogram(n_dft=1024, n_hop=128, input_shape=(1, dim),
                       padding='same', sr=dim, n_mels=80,
                       fmin=40.0, fmax=dim / 2, power_melgram=1.0,
                       return_decibel_melgram=True, trainable_fb=False,
                       trainable_kernel=False, trainable=False,
                       name='mel_stft')(x)
    spectrogram.trainable = False
    norm = norm_layer(spectrogram)

    x = layers.Permute((2, 1, 3))(norm)

    x = layers.Conv2D(10, (5, 1), activation='relu', padding='same')(x)
    x = layers.BatchNormalization()(x)
    x = layers.Conv2D(1, (5, 1), activation='relu', padding='same')(x)
    x = layers.BatchNormalization()(x)

    x = layers.Lambda(lambda q: K.squeeze(q, -1), name='squeeze_last_dim')(x)

    x = layers.Bidirectional(layers.LSTM(64, return_sequences=True))(x) 
    x = layers.Bidirectional(layers.LSTM(64, return_sequences=True))(x) 

    xFirst = layers.Lambda(lambda q: q[:, -1])(x)  # [b_s, vec_dim]
    query = layers.Dense(128)(xFirst)

    # dot product attention
    attScores = layers.Dot(axes=[1, 2])([query, x])
    attScores = layers.Softmax(name='attSoftmax')(attScores)  # [b_s, seq_len]

    # rescale sequence
    attVector = layers.Dot(axes=[1, 1])([attScores, x])  # [b_s, vec_dim]

    x = layers.Dense(64, activation='relu')(attVector)
    x = layers.Dense(32)(x)

    output = layers.Dense(num_labels, activation='softmax', name='output')(x)
    model =  keras.Model(inputs=[inputs], outputs=[output])
    model.summary()   
    return model
