import os
import pathlib
import wave
import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf
import math

seed = 42
tf.random.set_seed(seed)
np.random.seed(seed)


DATASET_PATH = 'data/mini_speech_commands'
data_dir = pathlib.Path(DATASET_PATH)
if not data_dir.exists():
  tf.keras.utils.get_file(
      'mini_speech_commands.zip',
      origin="http://storage.googleapis.com/download.tensorflow.org/data/mini_speech_commands.zip",
      extract=True,
      cache_dir='.', cache_subdir='data')

commands = np.array(tf.io.gfile.listdir(str(data_dir)))
commands = commands[commands != 'README.md']

AUTOTUNE = tf.data.experimental.AUTOTUNE

input_shape = (129,129)
sampling_matrix = tf.random.normal(shape=(input_shape[0], input_shape[1]), seed=1)

def get_size_dataset(train_ds):
  train_ds1 = train_ds.as_numpy_iterator()
  train_ds1 = list(train_ds1)
  return len(train_ds1)

def load_data_only_one():
    batch_size = 1
    train_data_size = 5300
    val_data_size = 1400
    test_data_size = 1300
    commands = np.array(tf.io.gfile.listdir(str(data_dir)))
    commands = commands[commands != 'README.md']

    filenames = np.loadtxt("filenames.txt", delimiter=';',  skiprows=1, dtype="str")
    filenames = [f[2:-1] for f in filenames]

    num_samples = len(filenames)
    
    train_files = filenames[:train_data_size]
    val_files = filenames[train_data_size: train_data_size + val_data_size]
    test_files = filenames[-test_data_size:]

    files_train = tf.data.Dataset.from_tensor_slices(train_files)
    waveform_train = files_train.map(map_func=get_waveform_and_label_id,num_parallel_calls=AUTOTUNE)
    waveform_train = waveform_train.filter(lambda x,y: tf.equal(tf.shape(x)[0], 16000))
    get_size_dataset(waveform_train)
    train_ds = waveform_train

    files_val = tf.data.Dataset.from_tensor_slices(val_files)
    waveform_val = files_val.map(map_func=get_waveform_and_label_id,num_parallel_calls=AUTOTUNE)
    waveform_val = waveform_val.filter(lambda x,y: tf.equal(tf.shape(x)[0], 16000))
    val_ds = waveform_val

    files_test = tf.data.Dataset.from_tensor_slices(test_files)
    waveform_test = files_test.map(map_func=get_waveform_and_label_id,num_parallel_calls=AUTOTUNE)
    waveform_test = waveform_test.filter(lambda x,y: tf.equal(tf.shape(x)[0], 16000))
    
    test_ds = waveform_test

    train_ds = train_ds.batch(batch_size)
    val_ds = val_ds.batch(batch_size)

    train_ds = train_ds.cache().prefetch(tf.data.experimental.AUTOTUNE)
    val_ds = val_ds.cache().prefetch(AUTOTUNE)
   
    for waveform, labels in waveform_train.take(1):
        input_shape = waveform.shape

    return train_ds, val_ds, test_ds, input_shape, waveform_train, commands


def load_data_batches():
    batch_size = 64
    train_data_size = 5300
    val_data_size = 1400
    test_data_size = 1300
    commands = np.array(tf.io.gfile.listdir(str(data_dir)))
    commands = commands[commands != 'README.md']
    
    filenames = np.loadtxt("filenames.txt", delimiter=';',  skiprows=1, dtype="str")
    filenames = [f[2:-1] for f in filenames]

    num_samples = len(filenames)
    
    train_files = filenames[:train_data_size]
    val_files = filenames[train_data_size: train_data_size + val_data_size]
    test_files = filenames[-test_data_size:]

    files_train = tf.data.Dataset.from_tensor_slices(train_files)
    waveform_train = files_train.map(map_func=get_waveform_and_label_id,num_parallel_calls=AUTOTUNE)
    waveform_train = waveform_train.filter(lambda x,y: tf.equal(tf.shape(x)[0], 16000))
    get_size_dataset(waveform_train)
    train_ds = waveform_train

    files_val = tf.data.Dataset.from_tensor_slices(val_files)
    waveform_val = files_val.map(map_func=get_waveform_and_label_id,num_parallel_calls=AUTOTUNE)
    waveform_val = waveform_val.filter(lambda x,y: tf.equal(tf.shape(x)[0], 16000))
    val_ds = waveform_val

    files_test = tf.data.Dataset.from_tensor_slices(test_files)
    waveform_test = files_test.map(map_func=get_waveform_and_label_id,num_parallel_calls=AUTOTUNE)
    waveform_test = waveform_test.filter(lambda x,y: tf.equal(tf.shape(x)[0], 16000))

    test_ds = waveform_test
    train_ds = train_ds.batch(batch_size)
    val_ds = val_ds.batch(batch_size)

    train_ds = train_ds.cache().prefetch(tf.data.experimental.AUTOTUNE)
    val_ds = val_ds.cache().prefetch(AUTOTUNE)
   
    for waveform, labels in waveform_train.take(1):
        input_shape = waveform.shape
    return train_ds, val_ds, test_ds, input_shape, waveform_train, commands

def decode_audio(audio_binary):
  audio, sample_rate = tf.audio.decode_wav(contents=audio_binary)
  return tf.squeeze(audio, axis=-1)

def get_label(file_path):
  parts = tf.strings.split(input=file_path, sep=os.path.sep)
  return parts[-2]

def get_waveform_and_label(file_path):
  label = get_label(file_path)
  audio_binary = tf.io.read_file(file_path)
  waveform = decode_audio(audio_binary)
  return waveform, label

def multiply_waveform(audio, label):
  audio = tf.math.multiply(audio, v2)
  return audio, label

def get_waveform_and_label_id(file_path):
  label = get_label(file_path)
  label_id = tf.argmax(label == commands)
  audio_binary = tf.io.read_file(file_path)
  waveform = decode_audio(audio_binary)
  return waveform, label_id

def get_spectrogram(waveform):
  waveform = tf.cast(waveform, dtype=tf.float32)
  spectrogram = tf.signal.stft(waveform, frame_length=255, frame_step=128)
  spectrogram = tf.abs(spectrogram)
  spectrogram = spectrogram[..., tf.newaxis]
  return spectrogram

def get_spectrogram_and_label_id(audio, label):
  spectrogram = get_spectrogram(audio)
  label_id = tf.argmax(label == commands)
  return spectrogram, label_id
  
def get_label_id(audio, label):
  label_id = tf.argmax(label == commands)
  return label_id

def make_spec_ds(ds):
  return ds.map(
      map_func=lambda audio,label: (get_spectrogram(audio), label),
      num_parallel_calls=tf.data.AUTOTUNE)

def preprocess_dataset(files):
  files_ds = tf.data.Dataset.from_tensor_slices(files)
  output_ds = files_ds.map(map_func=get_waveform_and_label,num_parallel_calls=AUTOTUNE)
  output_ds = output_ds.map(map_func=get_spectrogram_and_label_id, num_parallel_calls=AUTOTUNE)
  return output_ds

def get_spectrograms_from_files(files):
    files_ds = tf.data.Dataset.from_tensor_slices(files)
    waveform_ds = files_ds.map(map_func=get_waveform_and_label,num_parallel_calls=AUTOTUNE)
    spectrogram_ds = waveform_ds.map(map_func=get_spectrogram_and_label_id,num_parallel_calls=AUTOTUNE)
    return spectrogram_ds

def test_model(model, test_ds):
    test_audio = []
    test_labels = []
    for audio, label in test_ds:
        test_audio.append(audio.numpy())
        test_labels.append(label.numpy())
    test_audio = np.array(test_audio)
    test_labels = np.array(test_labels)  
    y_pred = np.argmax(model.predict(test_audio), axis=1)
    y_true = test_labels
    test_acc = sum(y_pred == y_true) / len(y_true)

  