import pathlib
import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf
from speech_functions import load_data_batches, get_size_dataset
from tensorflow.keras.callbacks import LearningRateScheduler
import pandas as pd
from speech_functions import *
from models import define_model, test_model

seed = 42
tf.random.set_seed(seed)
np.random.seed(seed)

def train_model(model, train_ds, val_ds):
    EPOCHS = 100
    early_stopping_monitor = tf.keras.callbacks.EarlyStopping(
        monitor='val_sparse_categorical_accuracy',
        min_delta=0,
        patience=10,
        verbose=0,
        mode='max',
        baseline=None,
        restore_best_weights=True
    )

    model_checkpoint_callback = tf.keras.callbacks.ModelCheckpoint(
        filepath='model_chk.h5',
        save_weights_only=True,
        monitor='val_sparse_categorical_accuracy',
        mode='max',
        save_best_only=True)

    def step_decay(epoch):
        initial_lrate = 0.001
        drop = 0.4
        epochs_drop = 15.0
        lrate = initial_lrate * math.pow(drop, math.floor((1+epoch)/epochs_drop))
        
        if (lrate < 4e-5):
            lrate = 4e-5

        return lrate

    learning_rate = LearningRateScheduler(step_decay)

    history = model.fit(
        train_ds,
        validation_data=val_ds,
        epochs=EPOCHS,
        callbacks=[model_checkpoint_callback, early_stopping_monitor, learning_rate], 
        verbose=2
    )
    model.load_weights('model_chk.h5')
    model.save("attention_model_sampling")

    plt.close('all')
    metrics = history.history
    plt.plot(history.epoch, metrics['loss'], metrics['val_loss'])
    plt.legend(['loss', 'val_loss'])
    plt.show()
    plt.savefig("training_loss_attention_sampling")
    return model

def compute_pruned_count(percent_pruning, layer):
    total = layer.output_shape[1]
    nb_to_be_pruned = int(np.floor(percent_pruning * total))
    return nb_to_be_pruned

def get_weights_l1(layer):
    weights = layer.get_weights()[0]
    norms = []
    for i in range(weights.shape[0]):
        l1_norm = np.sum(abs(weights[i]))
        norms.append(l1_norm)
    return np.array(norms)

def set_weights_to_zero(model, layer, layer_id, perc):  
    weights_norm = get_weights_l1(layer)
    weights = layer.get_weights()[0] 
    n_pruned = compute_pruned_count(perc, layer)
    low_norm_weights = weights_norm.argsort()[:n_pruned]
   
    for i in range(0, low_norm_weights.shape[0]):
        x = low_norm_weights[i]
        weights[x] = 0
    
    model.layers[layer_id].set_weights([weights])
    model.compile(
        optimizer=tf.keras.optimizers.Adam(learning_rate=0.001),
        loss=tf.keras.losses.SparseCategoricalCrossentropy(from_logits=False),
        metrics=['sparse_categorical_accuracy'], 
    )   
    return model

def test_trained_model():
    saved_model = tf.keras.models.load_model('attention_model_sampling') 
    saved_model.compile(
        optimizer=tf.keras.optimizers.Adam(learning_rate=0.001),
        loss=tf.keras.losses.SparseCategoricalCrossentropy(from_logits=False),
        metrics=['sparse_categorical_accuracy'], 
        )   
    test_model(saved_model)

    weights = saved_model.layers[1].get_weights()[0]
    np.savetxt("weights.csv", weights, delimiter=",")

    layer_id = 1
    for j in range(0, 10):
        perc = j/10
        model_pruned = set_weights_to_zero(saved_model, saved_model.layers[layer_id], layer_id, perc)
        test_model(model_pruned)
    return saved_model


def main():
    # Load data
    train_ds, val_ds, test_ds, input_shape, wave_orig, commands = load_data_batches()
    train_data_size = get_size_dataset(train_ds)
    num_labels = len(commands)

    model = define_model(num_labels=num_labels, input_shape=(16000,), dim=16000)
    model.compile(
        optimizer=tf.keras.optimizers.Adam(learning_rate=0.001),
        loss=tf.keras.losses.SparseCategoricalCrossentropy(from_logits=False),
        metrics=['sparse_categorical_accuracy'], 
    )     
    trained_model = train_model(model, train_ds, val_ds)
    test_model(trained_model, test_ds)

    #saved_model = test_trained_model()

