import pathlib
import matplotlib.pyplot as plt
import numpy as np
import tensorflow as tf
from speech_functions import load_data_batches
import pandas as pd
from models import define_sampling_subnet

seed = 42
tf.random.set_seed(seed)
np.random.seed(seed)

def compute_pruned_count(percent_pruning, layer):
    total = layer.output_shape[1]
    nb_to_be_pruned = int(np.floor(percent_pruning * total))
    return nb_to_be_pruned

def get_weights_l1(layer):
    weights = layer.get_weights()[0]
    norms = []
    for i in range(weights.shape[0]):
        l1_norm = np.sum(abs(weights[i]))
        norms.append(l1_norm)
    return np.array(norms)

def set_weights_to_zero(model, layer, layer_id, perc):  
    weights = layer.get_weights()[0] 
    weights_norm = abs(weights)
    n_pruned = compute_pruned_count(perc, layer)
    low_norm_weights = weights_norm.argsort()[:n_pruned]
    weights[low_norm_weights] = 0
    model.layers[layer_id].set_weights([weights])
    return model

def plot_predictions(model, val_ds):   
    weights = model.layers[1].get_weights()[0] 
    SMALL_SIZE = 14
    MEDIUM_SIZE = 16
    BIGGER_SIZE = 18

    plt.rc('font', size=SMALL_SIZE)          # controls default text sizes
    plt.rc('axes', titlesize=SMALL_SIZE)     # fontsize of the axes title
    plt.rc('axes', labelsize=MEDIUM_SIZE)    # fontsize of the x and y labels
    plt.rc('xtick', labelsize=SMALL_SIZE)    # fontsize of the tick labels
    plt.rc('ytick', labelsize=SMALL_SIZE)    # fontsize of the tick labels
    plt.rc('legend', fontsize=SMALL_SIZE)    # legend fontsize
    plt.rc('figure', titlesize=BIGGER_SIZE)  # fontsize of the figure title

    plt.clf()
    plt.plot(np.array(weights), color='red')
    plt.savefig("sampling_pattern")

    for i,(x, y) in enumerate(val_ds):  
        batch_predictions = model(x, training=False) 
        j = 0
        for inputs, samples in zip(x, batch_predictions):
            print(j)
            plt.clf()
            plt.plot(np.array(inputs), color='0.8', linestyle='solid', alpha=0.8, label='Raw waveform')
            plt.plot(np.array(samples), color='red', label='Subsampling pattern')
            plt.legend(loc="upper left")
            plt.xlabel("Sample index")
            plt.ylabel("Amplitude")
            plt.savefig("sampling_input"+str(j))
            j = j+1

def make_predictions(model, trainds):
    column_names = list(range(16000))
    for i in range(len(column_names)):
        column_names[i] = str(i+1)

    column_names2 = list(range(1))
    for i in range(len(column_names2)):
        column_names2[i] = str(i+1)
 
    labels = np.concatenate([np.expand_dims(y,axis=0) for x, y in trainds], axis=0)
    
    s = np.empty((0, 16000))
    for x, y in trainds:
        predictions = model(x, training=False)
        p = np.vstack((s, np.squeeze(np.array(predictions))))
        s = p

    dataframe1 = pd.DataFrame(labels, columns=column_names2)
    dataframe2 = pd.DataFrame(p, columns=column_names)

    dataframe1.to_csv('labels.csv', mode='a', header = False, index = None)
    dataframe2.to_csv('samples.csv', mode='a', header = False, index = None)


def main(path_to_pretrained_model):
    train_ds, val_ds, test_ds, input_shape, wave_orig, commands = load_data_batches()

    saved_model = tf.keras.models.load_model(path_to_pretrained_model) 
    saved_model.summary()
    weights = saved_model.layers[1].get_weights()[0]

    subsampling_model = define_sampling_subnet()

    l = []
    l.append(weights)
    subsampling_model.layers[1].set_weights(l)    
    layer_id = 1
    
    for x,y in train_ds:
        p = np.random.randint(0, 9) 
        perc = p/10
        model_pruned = set_weights_to_zero(subsampling_model, saved_model.layers[layer_id], layer_id, perc) 
        mini_train_dataset = tf.data.Dataset.from_tensor_slices((x, y))
        make_predictions(model_pruned, trainds=mini_train_dataset)



